/*******************************************************************************
*    IM-Daemon                                                                 *
*    Copyright (C) 2012-averyfarwaydate Luca Bertoncello                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <glib.h>
#include "incs/constants.h"
#include "incs/config.h"
#include "incs/logger.h"
#include "incs/options.h"
#include "incs/im.h"

char pidfile[255], configfile[255];
int purpleDebug;
struct config prgConfig;
GMainLoop *loop;
int fifoFD;

/**
 * Zeigt die Hilfe über das Programm
 */
void printHelp()
{
  printf("%s %s\n\n", NAME, VERSIONNR);
  printf("Usage:\n");
  printf("%s [options]:\n", NAME);
  printf("  --help, -h                         Display this help\n");
  printf("  --version, -v                      Display version number\n");
  printf("  --debug, -d                        Run in debug modus\n");
  printf("  --purpledebug, -p                  Enable internal debug of libpurple\n");
  printf("  --verbose, -V                      Increase verbosity\n");
  printf("  --quiet, -q                        Decrease verbosity\n");
  printf("  --logfile=<file>                   Logs in <file>. Default: %s\n", DEFAULTLOGFILE);
  printf("  --pidfile=<file>                   Save PID in <file>. Default: %s\n", DEFAULTPIDFILE);
  printf("  --config=<file>                    Configuration from <file>. Default: %s\n", DEFAULTCONFIG);
  exit(0);
}

/**
 * Beendet das Programm
 */
void quitFunc(int signal)
{
  LOGGER(LOGL_NOTICE, "Received signal %d. Terminate", signal);
  unlink(pidfile);
  g_main_loop_quit(loop);
}

/**
 * Schließt den Log und öffnet es neu.
 * Nützlich wegen logrotate
 */
void reloadFunc(int signal)
{
  closeLogger();
  openLogger();
  LOGGER(LOGL_INFO, "Received signal %d. Log subsystem reloaded", signal);
}

/**
 * Startet das Programm als Daemon
 */
void daemonize()
{
  pid_t pid, sid;
  FILE *fp;

  LOGGER(LOGL_NOTICE, "Starting daemon");

  pid = fork();
  if(pid < 0)
    exit(EXIT_FAILURE);

  if(pid > 0)
  {
    fp = fopen(pidfile, "wt");
    fprintf(fp, "%d", pid);
    fclose(fp);
    exit(EXIT_SUCCESS);
  }

  umask(0);

  sid = setsid();
  if(sid < 0)
    exit(EXIT_FAILURE);

  if((chdir("/")) < 0)
    exit(EXIT_FAILURE);

  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  close(STDERR_FILENO);
}

void readConfig()
{
  FILE *fp;
  char line[10240];
  char key[500], value[500];
  int c, l;
  char *protos[] = {"aim", "bonjour", "gg", "novell", "icq", "irc", "msn", "myspace",
                    "qq", "silc", "simple", "meanwhile", "jabber", "yahoo", "yahoojp", "zephyr"};

  bzero(&prgConfig, sizeof(prgConfig));
  fp = fopen(configfile, "rt");
  if(fp != NULL)
  {
    while(!feof(fp))
    {
      bzero(line, 10240);
      fgets(line, 10240, fp);
      if(!feof(fp))
      {
        while((line[strlen(line) - 1] == '\n') || (line[strlen(line) - 1] == '\r'))
          line[strlen(line) - 1] = 0;
        for(c = 0; c < strlen(line) && line[c] == ' '; c++)
          ;
        if(line[c] != '#' && strlen(line) != 0)
        {
          sscanf(line, "%s = %s", key, value);
          if(strcasecmp(key, "protocol") == 0)
          {
            for(l = 0; l < 16; l++)
              if(strcasecmp(value, protos[l]) == 0)
                strcpy(prgConfig.protocol, protos[l]);
            if(strlen(prgConfig.protocol) == 0)
              LOGGER(LOGL_FATAL, "Invalid protocol '%s'", value);
            if(strcmp(prgConfig.protocol, "icq") == 0)
            {
// Standardwerte für ICQ. Ansonsten hat libpurple noch alte Daten und loggt sich NICHT ein
              if(strlen(prgConfig.server) == 0)
                strcpy(prgConfig.server, "login.icq.com");
              if(prgConfig.port == 0)
                prgConfig.port = 5190;
            }
          }
          else if(strcasecmp(key, "login") == 0)
          {
            strcpy(prgConfig.login, value);
          }
          else if(strcasecmp(key, "password") == 0)
          {
            strcpy(prgConfig.password, value);
          }
          else if(strcasecmp(key, "server") == 0)
          {
            strcpy(prgConfig.server, value);
          }
          else if(strcasecmp(key, "port") == 0)
          {
            prgConfig.port = atoi(value);
            if(prgConfig.port < 1 || prgConfig.port > 65535)
              LOGGER(LOGL_FATAL, "Invalid port number %s", value);
          }
          else if(strcasecmp(key, "requireTLS") == 0)
          {
            if(strcasecmp(value, "true") == 0)
              prgConfig.requireTLS = TRUE;
            else if(strcasecmp(value, "false") == 0)
              prgConfig.requireTLS = FALSE;
            else
              LOGGER(LOGL_ERROR, "Unknown value '%s' for key '%s'. Using default (false)", value, key);
          }
          else if(strcasecmp(key, "iconpath") == 0)
          {
            strcpy(prgConfig.iconPath, value);
          }
          else if(strcasecmp(key, "fifo") == 0)
          {
            strcpy(prgConfig.fifoPath, value);
          }
          else
            LOGGER(LOGL_ERROR, "Unknown key '%s'. Ignoring...", key);
        }
      }
    }
    fclose(fp);
  }
  else
    LOGGER(LOGL_FATAL, "Unable to read configuration file '%s'", configfile);
}

/**
 * Hauptschleife.
 * Alle Meldungen werden bearbeitet.
 * Wenn ein  Stopsignal kommt, wird die Schleife beendet
 *
 * @param gpointer data   Zeiger zum GMainLoop-Struktur
 * @return gboolean       TRUE wenn alles in Ordnung, FALSE, wenn die Schleife beendet werden soll
 */
gboolean mainLoopCB(gpointer data)
{
  char line[10240], recipient[255], message[10240], *p;
  ssize_t r;
  static time_t reconnectWait;

// Wenn die Verbindung getrennt worden ist, versucht sich erneut zu verbinden
  if(!isSignedOn() && !isConnecting())
  {
    if(reconnectWait == 0)
    {
      reconnectWait = time(NULL);
      LOGGER(LOGL_INFO, "Waiting %d seconds before reconnect", DELAYFORRECONNECT);
    }
    else
    {
      if(time(NULL) > reconnectWait + DELAYFORRECONNECT)
      {
        reconnectWait = 0;
        connectToIMServer(prgConfig);
      }
    }
  }
  else
    reconnectWait = 0;

  bzero(line, 10240);
  r = read(fifoFD, line, 10240);
  if(r > 0)
  {
    while(line[strlen(line) - 1] == '\n' || line[strlen(line) - 1] == '\r')
      line[strlen(line) - 1] = 0;
    bzero(recipient, 255);
    bzero(message, 10240);
    p = index(line, '\t');
    if(p != NULL)
    {
      strcpy(message, p + 1);
      *p = 0;
      strcpy(recipient, line);
      sendIM(recipient, message);
    }
    else
      LOGGER(LOGL_NOTICE, "Invalid line received. Please send in the format 'recipient<TAB>message'");
  }

  return TRUE;
}

int main(int argc, char* argv[])
{
  struct options opts;
  gboolean help, version, verbose, quiet;
  int l, debug;
  char defLog[255], logfile[255], defPid[255], defConf[255];
  struct sigaction act, reload;

  strcpy(defLog, DEFAULTLOGFILE);
  strcpy(defPid, DEFAULTPIDFILE);
  strcpy(defConf, DEFAULTCONFIG);

  opts.nOptions = 9;
  opts.invalidOptionMsg = malloc(sizeof(char) * 255);
  strcpy(opts.invalidOptionMsg, "Invalid option [%s]\n");
  opts.opt = malloc(sizeof(struct option) * opts.nOptions);
  for(l = 0; l < opts.nOptions; l++)
    opts.opt[l].optNameLong = malloc(sizeof(char) * 20);
  l = debug = purpleDebug = verbose = quiet = 0;
  strcpy(opts.opt[l].optNameLong, "logfile");
  opts.opt[l].optNameShort = 0;
  opts.opt[l].type = STRING;
  opts.opt[l].value = logfile;
  opts.opt[l].defValue = defLog;
  strcpy(opts.opt[++l].optNameLong, "pidfile");
  opts.opt[l].optNameShort = 0;
  opts.opt[l].type = STRING;
  opts.opt[l].value = pidfile;
  opts.opt[l].defValue = defPid;
  strcpy(opts.opt[++l].optNameLong, "config");
  opts.opt[l].optNameShort = 0;
  opts.opt[l].type = STRING;
  opts.opt[l].value = configfile;
  opts.opt[l].defValue = defConf;
  strcpy(opts.opt[++l].optNameLong, "help");
  opts.opt[l].optNameShort = 'h';
  opts.opt[l].type = BOOL;
  opts.opt[l].value = &help;
  opts.opt[l].defValue = NULL;
  strcpy(opts.opt[++l].optNameLong, "version");
  opts.opt[l].optNameShort = 'v';
  opts.opt[l].type = BOOL;
  opts.opt[l].value = &version;
  opts.opt[l].defValue = NULL;
  strcpy(opts.opt[++l].optNameLong, "debug");
  opts.opt[l].optNameShort = 'd';
  opts.opt[l].type = BOOL;
  opts.opt[l].value = &debug;
  opts.opt[l].defValue = NULL;
  strcpy(opts.opt[++l].optNameLong, "purpledebug");
  opts.opt[l].optNameShort = 'p';
  opts.opt[l].type = BOOL;
  opts.opt[l].value = &purpleDebug;
  opts.opt[l].defValue = NULL;
  strcpy(opts.opt[++l].optNameLong, "verbose");
  opts.opt[l].optNameShort = 'V';
  opts.opt[l].type = BOOL;
  opts.opt[l].value = &verbose;
  opts.opt[l].defValue = NULL;
  strcpy(opts.opt[++l].optNameLong, "quiet");
  opts.opt[l].optNameShort = 'q';
  opts.opt[l].type = BOOL;
  opts.opt[l].value = &quiet;
  opts.opt[l].defValue = NULL;

  if((getOptions(argc, argv, opts) != 0 || help != 0) && version == 0)
    printHelp();

  if(version == 1)
  {
    printf("%s %s\n", NAME, VERSIONNR);
    exit(EXIT_SUCCESS);
  }

  free(opts.invalidOptionMsg);
  for(l = 0; l < opts.nOptions; l++)
    free(opts.opt[l].optNameLong);
  free(opts.opt);

  setLogFile(logfile);
  if(debug)
    setLogLevel(LOGL_DEBUG);
  else
  {
    if(verbose)
      setLogLevel(LOGL_INFO);
    else if(quiet)
      setLogLevel(LOGL_ERROR);
    else
      setLogLevel(LOGL_NOTICE);
  }

  openLogger();

  readConfig();
  if(strlen(prgConfig.protocol) == 0)
    LOGGER(LOGL_FATAL, "Protocol is mandatory");
  if(strlen(prgConfig.login) == 0)
    LOGGER(LOGL_FATAL, "Login is mandatory");
  if(strlen(prgConfig.password) == 0)
    LOGGER(LOGL_FATAL, "Password is mandatory");
  if(strlen(prgConfig.fifoPath) == 0)
    LOGGER(LOGL_FATAL, "FIFO-Path is mandatory");

  act.sa_handler = quitFunc;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if(sigaction(SIGQUIT, &act, NULL) != 0)
    LOGGER(LOGL_FATAL, "Unable to set catch function for signal");
  if(sigaction(SIGINT, &act, NULL) != 0)
    LOGGER(LOGL_FATAL, "Unable to set catch function for signal");
  if(sigaction(SIGTERM, &act, NULL) != 0)
    LOGGER(LOGL_FATAL, "Unable to set catch function for signal");
  reload.sa_handler = reloadFunc;
  sigemptyset(&reload.sa_mask);
  reload.sa_flags = 0;
  if(sigaction(SIGHUP, &reload, NULL) != 0)
    LOGGER(LOGL_FATAL, "Unable to set catch function for signal");

  if(mkfifo(prgConfig.fifoPath, 0666) != 0)
    LOGGER(LOGL_FATAL, "Unable to create FIFO %s", prgConfig.fifoPath);
  chmod(prgConfig.fifoPath, 0666);
  if((fifoFD = open(prgConfig.fifoPath, O_RDONLY | O_NONBLOCK)) == -1)
  {
    unlink(prgConfig.fifoPath);
    LOGGER(LOGL_FATAL, "Unable to open FIFO %s", prgConfig.fifoPath);
  }

  if(!debug)
    daemonize();

  initPurple();
  connectToIMServer(prgConfig);

  loop = g_main_loop_new(NULL, FALSE);
  g_timeout_add(100, mainLoopCB, loop);

  LOGGER(LOGL_INFO, "Starting main loop");
  g_main_loop_run(loop);

  disconnectFromIMServer();

  close(fifoFD);
  unlink(prgConfig.fifoPath);
  closeLogger();

  return EXIT_SUCCESS;
}
